@@ -1,16 +1,18 @@
+use std::fmt::Write;
 use std::sync::mpsc::channel;
 use std::time::Instant;
-use std::unimplemented;
 
+use crate::output::CharacterChunk;
 use crate::panes::PaneId;
 use crate::pty::VteBytes;
 use crate::tab::Pane;
-use crate::ui::pane_boundaries_frame::PaneFrame;
+use crate::ui::pane_boundaries_frame::{FrameParams, PaneFrame};
 use crate::wasm_vm::PluginInstruction;
+use crate::ClientId;
 use zellij_utils::pane_size::Offset;
 use zellij_utils::position::Position;
 use zellij_utils::shared::ansi_len;
-use zellij_utils::zellij_tile::prelude::{Event, Mouse, PaletteColor};
+use zellij_utils::zellij_tile::prelude::{Event, InputMode, Mouse, PaletteColor};
 use zellij_utils::{
     channels::SenderWithContext,
     pane_size::{Dimension, PaneGeom},
@@ -26,8 +28,8 @@ pub(crate) struct PluginPane {
     pub send_plugin_instructions: SenderWithContext<PluginInstruction>,
     pub active_at: Instant,
     pub pane_title: String,
+    pub pane_name: String,
     frame: bool,
-    frame_color: Option<PaletteColor>,
     borderless: bool,
 }
 
@@ -37,6 +39,7 @@ impl PluginPane {
         position_and_size: PaneGeom,
         send_plugin_instructions: SenderWithContext<PluginInstruction>,
         title: String,
+        pane_name: String,
     ) -> Self {
         Self {
             pid,
@@ -47,10 +50,10 @@ impl PluginPane {
             send_plugin_instructions,
             active_at: Instant::now(),
             frame: false,
-            frame_color: None,
             content_offset: Offset::default(),
             pane_title: title,
             borderless: false,
+            pane_name,
         }
     }
 }
@@ -102,13 +105,14 @@ impl Pane for PluginPane {
         self.should_render = true;
     }
     fn handle_pty_bytes(&mut self, _event: VteBytes) {
-        unimplemented!()
+        // noop
     }
     fn cursor_coordinates(&self) -> Option<(usize, usize)> {
         None
     }
     fn adjust_input_to_terminal(&self, _input_bytes: Vec<u8>) -> Vec<u8> {
-        unimplemented!()
+        // noop
+        vec![]
     }
     fn position_and_size(&self) -> PaneGeom {
         self.geom
@@ -131,7 +135,13 @@ impl Pane for PluginPane {
     fn set_selectable(&mut self, selectable: bool) {
         self.selectable = selectable;
     }
-    fn render(&mut self) -> Option<String> {
+    fn render(
+        &mut self,
+        client_id: Option<ClientId>,
+    ) -> Option<(Vec<CharacterChunk>, Option<String>)> {
+        // this is a bit of a hack but works in a pinch
+        client_id?;
+        let client_id = client_id.unwrap();
         // if self.should_render {
         if true {
             // while checking should_render rather than rendering each pane every time
@@ -145,6 +155,7 @@ impl Pane for PluginPane {
                 .send(PluginInstruction::Render(
                     buf_tx,
                     self.pid,
+                    client_id,
                     self.get_content_rows(),
                     self.get_content_columns(),
                 ))
@@ -152,17 +163,6 @@ impl Pane for PluginPane {
 
             self.should_render = false;
             let contents = buf_rx.recv().unwrap();
-            // FIXME: This is a hack that assumes all fixed-size panes are borderless. This
-            // will eventually need fixing!
-            if self.frame && !(self.geom.rows.is_fixed() || self.geom.cols.is_fixed()) {
-                let frame = PaneFrame {
-                    geom: self.current_geom().into(),
-                    title: self.pane_title.clone(),
-                    color: self.frame_color,
-                    ..Default::default()
-                };
-                vte_output.push_str(&frame.render());
-            }
             for (index, line) in contents.lines().enumerate() {
                 let actual_len = ansi_len(line);
                 let line_to_print = if actual_len > self.get_content_columns() {
@@ -177,12 +177,14 @@ impl Pane for PluginPane {
                     .concat()
                 };
 
-                vte_output.push_str(&format!(
+                write!(
+                    &mut vte_output,
                     "\u{1b}[{};{}H\u{1b}[m{}",
                     self.get_content_y() + 1 + index,
                     self.get_content_x() + 1,
                     line_to_print,
-                )); // goto row/col and reset styles
+                )
+                .unwrap(); // goto row/col and reset styles
                 let line_len = line_to_print.len();
                 if line_len < self.get_content_columns() {
                     // pad line
@@ -197,21 +199,74 @@ impl Pane for PluginPane {
                 for line_index in total_line_count..self.get_content_rows() {
                     let x = self.get_content_x();
                     let y = self.get_content_y();
-                    vte_output.push_str(&format!(
+                    write!(
+                        &mut vte_output,
                         "\u{1b}[{};{}H\u{1b}[m",
                         y + line_index + 1,
                         x + 1
-                    )); // goto row/col and reset styles
+                    )
+                    .unwrap(); // goto row/col and reset styles
                     for _col_index in 0..self.get_content_columns() {
                         vte_output.push(' ');
                     }
                 }
             }
-            Some(vte_output)
+            Some((vec![], Some(vte_output))) // TODO: PluginPanes should have their own grid so that we can return the non-serialized TerminalCharacters and have them participate in the render buffer
         } else {
             None
         }
     }
+    fn render_frame(
+        &mut self,
+        _client_id: ClientId,
+        frame_params: FrameParams,
+        input_mode: InputMode,
+    ) -> Option<(Vec<CharacterChunk>, Option<String>)> {
+        // FIXME: This is a hack that assumes all fixed-size panes are borderless. This
+        // will eventually need fixing!
+        if self.frame && !(self.geom.rows.is_fixed() || self.geom.cols.is_fixed()) {
+            let pane_title = if self.pane_name.is_empty()
+                && input_mode == InputMode::RenamePane
+                && frame_params.is_main_client
+            {
+                String::from("Enter name...")
+            } else if self.pane_name.is_empty() {
+                self.pane_title.clone()
+            } else {
+                self.pane_name.clone()
+            };
+            let frame = PaneFrame::new(
+                self.current_geom().into(),
+                (0, 0), // scroll position
+                pane_title,
+                frame_params,
+            );
+            Some(frame.render())
+        } else {
+            None
+        }
+    }
+    fn render_fake_cursor(
+        &mut self,
+        _cursor_color: PaletteColor,
+        _text_color: PaletteColor,
+    ) -> Option<String> {
+        None
+    }
+    fn update_name(&mut self, name: &str) {
+        match name {
+            "\0" => {
+                self.pane_name = String::new();
+            }
+            "\u{007F}" | "\u{0008}" => {
+                //delete and backspace keys
+                self.pane_name.pop();
+            }
+            c => {
+                self.pane_name.push_str(c);
+            }
+        }
+    }
     fn pid(&self) -> PaneId {
         PaneId::Plugin(self.pid)
     }
@@ -255,45 +310,50 @@ impl Pane for PluginPane {
         self.geom.y -= count;
         self.should_render = true;
     }
-    fn scroll_up(&mut self, count: usize) {
+    fn scroll_up(&mut self, count: usize, client_id: ClientId) {
         self.send_plugin_instructions
             .send(PluginInstruction::Update(
                 Some(self.pid),
+                Some(client_id),
                 Event::Mouse(Mouse::ScrollUp(count)),
             ))
             .unwrap();
     }
-    fn scroll_down(&mut self, count: usize) {
+    fn scroll_down(&mut self, count: usize, client_id: ClientId) {
         self.send_plugin_instructions
             .send(PluginInstruction::Update(
                 Some(self.pid),
+                Some(client_id),
                 Event::Mouse(Mouse::ScrollDown(count)),
             ))
             .unwrap();
     }
     fn clear_scroll(&mut self) {
-        unimplemented!();
+        // noop
     }
-    fn start_selection(&mut self, start: &Position) {
+    fn start_selection(&mut self, start: &Position, client_id: ClientId) {
         self.send_plugin_instructions
             .send(PluginInstruction::Update(
                 Some(self.pid),
+                Some(client_id),
                 Event::Mouse(Mouse::LeftClick(start.line.0, start.column.0)),
             ))
             .unwrap();
     }
-    fn update_selection(&mut self, position: &Position) {
+    fn update_selection(&mut self, position: &Position, client_id: ClientId) {
         self.send_plugin_instructions
             .send(PluginInstruction::Update(
                 Some(self.pid),
+                Some(client_id),
                 Event::Mouse(Mouse::Hold(position.line.0, position.column.0)),
             ))
             .unwrap();
     }
-    fn end_selection(&mut self, end: Option<&Position>) {
+    fn end_selection(&mut self, end: Option<&Position>, client_id: ClientId) {
         self.send_plugin_instructions
             .send(PluginInstruction::Update(
                 Some(self.pid),
+                Some(client_id),
                 Event::Mouse(Mouse::Release(
                     end.map(|Position { line, column }| (line.0, column.0)),
                 )),
@@ -317,14 +377,19 @@ impl Pane for PluginPane {
     fn set_content_offset(&mut self, offset: Offset) {
         self.content_offset = offset;
     }
-    fn set_boundary_color(&mut self, color: Option<PaletteColor>) {
-        self.frame_color = color;
-        self.set_should_render(true);
-    }
     fn set_borderless(&mut self, borderless: bool) {
         self.borderless = borderless;
     }
     fn borderless(&self) -> bool {
         self.borderless
     }
+    fn handle_right_click(&mut self, to: &Position, client_id: ClientId) {
+        self.send_plugin_instructions
+            .send(PluginInstruction::Update(
+                Some(self.pid),
+                Some(client_id),
+                Event::Mouse(Mouse::RightClick(to.line.0, to.column.0)),
+            ))
+            .unwrap();
+    }
 }
